@@ -9,7 +9,7 @@ import (
 
 // change these to test a different UART or pins if available
 var (
-	uart = machine.UART0
+	uart = machine.Serial
 	tx   = machine.UART_TX_PIN
 	rx   = machine.UART_RX_PIN
 )
