@@ -1,5 +1,4 @@
 //go:build rp2040
-// +build rp2040
 
 package machine
 
@@ -42,8 +41,12 @@ func (uart *UART) Configure(config UARTConfig) error {
 		rp.UART0_UARTCR_TXE)
 
 	// set GPIO mux to UART for the pins
-	config.TX.Configure(PinConfig{Mode: PinUART})
-	config.RX.Configure(PinConfig{Mode: PinUART})
+	if config.TX != NoPin {
+		config.TX.Configure(PinConfig{Mode: PinUART})
+	}
+	if config.RX != NoPin {
+		config.RX.Configure(PinConfig{Mode: PinUART})
+	}
 
 	// Enable RX IRQ.
 	uart.Interrupt.SetPriority(0x80)
