@@ -1,4 +1,4 @@
-// +build darwin linux,!baremetal
+//go:build darwin || (linux && !baremetal)
 
 // Copyright 2016 The Go Authors. All rights reserved.
 // Use of this source code is governed by a BSD-style
@@ -10,11 +10,6 @@ import (
 	"syscall"
 )
 
-// Sync is a stub, not yet implemented
-func (f *File) Sync() error {
-	return ErrNotImplemented
-}
-
 // Stat returns the FileInfo structure describing file.
 // If there is an error, it will be of type *PathError.
 func (f *File) Stat() (FileInfo, error) {
