@@ -1,5 +1,4 @@
 //go:build darwin
-// +build darwin
 
 package syscall
 
@@ -14,10 +13,10 @@ import (
 // This function returns the error location in the darwin ABI.
 // Discovered by compiling the following code using Clang:
 //
-//     #include <errno.h>
-//     int getErrno() {
-//         return errno;
-//     }
+//	#include <errno.h>
+//	int getErrno() {
+//	    return errno;
+//	}
 //
 //export __error
 func libc___error() *int32
@@ -53,6 +52,9 @@ const (
 	DT_SOCK    = 0xc
 	DT_UNKNOWN = 0x0
 	DT_WHT     = 0xe
+	F_GETFL    = 0x3
+	F_SETFL    = 0x4
+	O_NONBLOCK = 0x4
 )
 
 // Source: https://opensource.apple.com/source/xnu/xnu-7195.81.3/bsd/sys/errno.h.auto.html
@@ -63,10 +65,12 @@ const (
 	EEXIST      Errno = 17
 	EINTR       Errno = 4
 	ENOTDIR     Errno = 20
+	EISDIR      Errno = 21
 	EINVAL      Errno = 22
 	EMFILE      Errno = 24
 	EPIPE       Errno = 32
 	EAGAIN      Errno = 35
+	ENOTCONN    Errno = 57
 	ETIMEDOUT   Errno = 60
 	ENOSYS      Errno = 78
 	EWOULDBLOCK Errno = EAGAIN
@@ -143,6 +147,11 @@ type Timespec struct {
 	Nsec int64
 }
 
+// Unix returns the time stored in ts as seconds plus nanoseconds.
+func (ts *Timespec) Unix() (sec int64, nsec int64) {
+	return int64(ts.Sec), int64(ts.Nsec)
+}
+
 // Source: upstream ztypes_darwin_amd64.go
 type Dirent struct {
 	Ino       uint64
@@ -154,7 +163,6 @@ type Dirent struct {
 	Pad_cgo_0 [3]byte
 }
 
-// Go chose Linux's field names for Stat_t, see https://github.com/golang/go/issues/31735
 type Stat_t struct {
 	Dev       int32
 	Mode      uint16
@@ -164,10 +172,10 @@ type Stat_t struct {
 	Gid       uint32
 	Rdev      int32
 	Pad_cgo_0 [4]byte
-	Atim      Timespec
-	Mtim      Timespec
-	Ctim      Timespec
-	Btim      Timespec
+	Atimespec Timespec
+	Mtimespec Timespec
+	Ctimespec Timespec
+	Btimespec Timespec
 	Size      int64
 	Blocks    int64
 	Blksize   int32
@@ -259,6 +267,23 @@ func Pipe2(fds []int, flags int) (err error) {
 	return
 }
 
+func Chmod(path string, mode uint32) (err error) {
+	data := cstring(path)
+	fail := int(libc_chmod(&data[0], mode))
+	if fail < 0 {
+		err = getErrno()
+	}
+	return
+}
+
+func closedir(dir uintptr) (err error) {
+	e := libc_closedir(unsafe.Pointer(dir))
+	if e != 0 {
+		err = getErrno()
+	}
+	return
+}
+
 func readdir_r(dir uintptr, entry *Dirent, result **Dirent) (err error) {
 	e1 := libc_readdir_r(unsafe.Pointer(dir), unsafe.Pointer(entry), unsafe.Pointer(result))
 	if e1 != 0 {
@@ -267,31 +292,47 @@ func readdir_r(dir uintptr, entry *Dirent, result **Dirent) (err error) {
 	return
 }
 
-// The odd $INODE64 suffix is an Apple compatibility feature, see
-// https://assert.cc/posts/darwin_use_64_bit_inode_vs_ctypes/
-// and https://github.com/golang/go/issues/35269
-// Without it, you get the old, smaller struct stat from mac os 10.2 or so.
+func Getpagesize() int {
+	return int(libc_getpagesize())
+}
 
-// struct DIR * buf fdopendir(int fd);
-//export fdopendir$INODE64
-func libc_fdopendir(fd int32) unsafe.Pointer
+// The following RawSockAddr* types have been copied from the Go source tree and
+// are here purely to fix build errors.
 
-// int readdir_r(struct DIR * buf, struct dirent *entry, struct dirent **result);
-//export readdir_r$INODE64
-func libc_readdir_r(unsafe.Pointer, unsafe.Pointer, unsafe.Pointer) int32
+type RawSockaddr struct {
+	Len    uint8
+	Family uint8
+	Data   [14]int8
+}
 
-// int stat(const char *path, struct stat * buf);
-//export stat$INODE64
-func libc_stat(pathname *byte, ptr unsafe.Pointer) int32
+type RawSockaddrInet4 struct {
+	Len    uint8
+	Family uint8
+	Port   uint16
+	Addr   [4]byte /* in_addr */
+	Zero   [8]int8
+}
 
-// int fstat(int fd, struct stat * buf);
-//export fstat$INODE64
-func libc_fstat(fd int32, ptr unsafe.Pointer) int32
-
-// int lstat(const char *path, struct stat * buf);
-//export lstat$INODE64
-func libc_lstat(pathname *byte, ptr unsafe.Pointer) int32
+type RawSockaddrInet6 struct {
+	Len      uint8
+	Family   uint8
+	Port     uint16
+	Flowinfo uint32
+	Addr     [16]byte /* in6_addr */
+	Scope_id uint32
+}
 
 // int pipe(int32 *fds);
+//
 //export pipe
 func libc_pipe(fds *int32) int32
+
+// int getpagesize();
+//
+//export getpagesize
+func libc_getpagesize() int32
+
+// int open(const char *pathname, int flags, mode_t mode);
+//
+//export syscall_libc_open
+func libc_open(pathname *byte, flags int32, mode uint32) int32
